/*------------------------------------------------------------------------
   Andrew Hamilton 21 Feb 2001
   Andrew.Hamilton@colorado.edu
   http://casa.colorado.edu/~ajsh

   Find roots x0, xm, xp of cubic
         x^3 + a x^2 + b x + c = 0
   If all roots are real, they are ordered x0 <= xm <= xp.
   If two roots are complex, they are xm +- I xp, and the real root is x0.

   Compile with -DDOUBLE to treat REAL as double.

   This routine is intended to be accurate and robust.

   Synopsis:

#include "cubic.h"

    REAL a, b, c;
    roots3 rt;

    ier = cubic(a, b, c, &rt);

------------------------------------------------------------------------*/

#include <stdio.h>
#include "cubic.h"

/* tolerance within which to recognize almost degenerate roots as equal */
#ifdef DOUBLE
# define CHOP	1.e-12
#else
# define CHOP	1.e-6
#endif

/*------------------------------------------------------------------------
   Roots x0, xm, xp of cubic
      x^3 + a x^2 + b x + c = 0
   If all roots are real, they are ordered x0 <= xm <= xp.
   If two roots are complex, they are xm +- I xp.

   Return: 0 = ok
	   1 = error, which should never happen.
*/
int cubic(a, b, c, rt)
REAL a, b, c;
roots3 *rt;
{
    int sr;
    REAL alpha, af, ap, aq, ar, bf, cf, d, f, p, q, r, c3, sq, s3;

#ifndef NOIEEE
    /* one or more of a b c is not finite */
    if (!finite(a) || !finite(b) || !finite(c)) {
	return(cubic_oo(a, b, c, rt));
    }
#endif

    /* c = 0 warrants special treatment */
    if (c == 0.) {
	if (b == 0.) {				/* b = 0 */
	    if (a == 0.) {			/* a = 0 */
		rt->type = TRIPLE;
		rt->x0 = rt->xm = rt->xp = 0.;
	    } else if (a > 0.) {		/* a > 0 */
		rt->type = UPPER_DOUBLE;
		rt->x0 = - a;
		rt->xm = rt->xp = 0.;
	    } else {				/* a < 0 */
		rt->type = LOWER_DOUBLE;
		rt->x0 = rt->xm = 0.;
		rt->xp = - a;
	    }
	} else {				/* b != 0 */
	    /* renormalize coefficients so largest is +- 1
	       both to avoid under/overflow,
	       and to ensure consistent identification of degenerate roots */
	    af = fabs(a);
	    bf = sqrt(fabs(b));
	    f = (af > bf)? af : bf;
	    af = a / f;
	    bf = b / f / f;
	    /* discriminant */
	    q = af * af - 4. * bf;
	    aq = (q >= 0.)? q : -q;
	    /* 2 degenerate roots */
	    if (aq <= CHOP) {
		if (a > 0.) {			/* a > 0 */
		    rt->type = LOWER_DOUBLE;
		    rt->x0 = rt->xm = - a / 2.;
		    rt->xp = 0.;
		} else if (a < 0.) {		/* a < 0 */
		    rt->type = UPPER_DOUBLE;
		    rt->x0 = 0.;
		    rt->xm = rt->xp = - a / 2.;
		} else {			/* a = 0 */
		    rt->type = NONE;
		    printf(" cubic: this should not happen 1; a b c = %lf %lf %lf\n", a, b, c);
		    return(1);
		}
	    /* 2 real roots */
	    } else if (q > 0.) {
		rt->type = THREE;
		sq = sqrt(q);
		if (a > 0.) {			/* a > 0 */
		    p = - af - sq;
		    rt->x0 = p / 2. * f;
		    if (b > 0.) {		/* b > 0 */
			rt->xm = 2. * bf / p * f;
			rt->xp = 0.;
		    } else if (b < 0.) {	/* b < 0 */
			rt->xm = 0.;
			rt->xp = 2. * bf / p * f;
		    } else {			/* b = 0 */
			rt->type = NONE;
			printf(" cubic: this should not happen 2; a b c = %lf %lf %lf\n", a, b, c);
			return(1);
		    }
		} else if (a < 0.) {		/* a < 0 */
		    p = - af + sq;
		    rt->xp = p / 2. * f;
		    if (b > 0.) {		/* b > 0 */
			rt->x0 = 0.;
			rt->xm = 2. * bf / p * f;
		    } else if (b < 0.) {	/* b < 0 */
			rt->x0 = 2. * bf / p * f;
			rt->xm = 0.;
		    } else {			/* b = 0 */
			rt->type = NONE;
			printf(" cubic: this should not happen 3; a b c = %lf %lf %lf\n", a, b, c);
			return(1);
		    }
		} else {			/* a = 0 */
		    rt->xm = 0.;
		    rt->xp = sq / 2. * f;
		    rt->x0 = - rt->xp;
		}
	    /* 2 complex conjugate roots */
	    } else if (q < 0.) {
		rt->type = ONE;
		sq = sqrt(-q);
		rt->x0 = 0.;
		rt->xm = - a / 2.;
		rt->xp = sq / 2. * f;
	    } else {
		rt->type = NONE;
		printf(" cubic: this should not happen 4; a b c = %lf %lf %lf\n", a, b, c);
		return(1);
	    }
	}
	return(0);
    }

    /* renormalize coefficients so largest is +- 1
       both to avoid under/overflow,
       and to ensure consistent identification of degenerate roots */
    af = fabs(a);
    bf = sqrt(fabs(b));
    cf = pow(fabs(c), 1./3.);
    f = (af > bf)? af : bf;
    f = (f > cf)? f : cf;
    if (f > 0.) {
	af = a / f;
	bf = b / f / f;
	cf = c / f / f / f;
    } else {
	f = 1.;
	af = a;
	bf = b;
	cf = c;
    }

    /* discriminants */
    q = af * af - 3. * bf;
    d = 1.5 * (9. * cf - af * bf);
    r = af * q + d;
    /* p = q^3 - r^2 */
    p = 9. * q * (bf * bf - 3. * af * cf) - d * d;

    aq = (q >= 0.)? q : -q;
    ar = (r >= 0.)? r : -r;
    ap = (p >= 0.)? p : -p;
    sr = (r >= 0.)? 1 : -1;
    af = - af / 3.;
    sq = sqrt(aq);
    d = sr * sq / 3.;

    /* 1 degenerate triple root */
    if (aq <= CHOP && ar <= CHOP) {
	rt->type = TRIPLE;
	/* x0 = xm = xp */
	rt->x0 = rt->xm = rt->xp = af * f;
    /* 2 real roots, including 1 degenerate root */
    } else if (q > 0. && ap <= CHOP * r * r) {
        if (sr > 0) {
            rt->type = UPPER_DOUBLE;
            /* x0 < xm = xp */
            rt->x0 = (af - 2. * d) * f;
            rt->xm = rt->xp = (af + d) * f;
        } else {
            rt->type = LOWER_DOUBLE;
            /* x0 = xm < xp */
            rt->x0 = rt->xm = (af + d) * f;
            rt->xp = (af - 2. * d) * f;
        }
    /* 3 nondegenerate real roots */
    } else if (q > 0. && p >= 0.) {
	rt->type = THREE;
	alpha = atan(sqrt(ap) / ar);
	c3 = cos(alpha / 3.);
	s3 = SQRT3 * sin(alpha / 3.);
	/* x0 < xm < xp */
	if (sr > 0) {
	    rt->x0 = (af - 2. * c3 * d) * f;
	    rt->xm = (af + (c3 - s3) * d) * f;
	    rt->xp = (af + (c3 + s3) * d) * f;
	} else {
	    rt->xp = (af - 2. * c3 * d) * f;
	    rt->xm = (af + (c3 - s3) * d) * f;
	    rt->x0 = (af + (c3 + s3) * d) * f;
	}
    /* 1 nondegenerate real root */
    } else {				/* q < 0 || p < 0 */
	rt->type = ONE;
	if (aq <= ERR_TIGHT) {		/* q = 0 */
	    d = sr * pow(ar / 4., 1./3.) / 3.;
	    rt->x0 = (af - 2. * d) * f;
	    rt->xm = (af + d) * f;
	    rt->xp = (sr * SQRT3 * d) * f;
	} else if (q > 0.) {		/* q > 0 && p < 0 */
	    alpha = ASINH(sqrt(ap) / sq / aq);
	    c3 = cosh(alpha / 3.);
	    s3 = SQRT3 * sinh(alpha / 3.);
	    rt->x0 = (af - 2. * c3 * d) * f;
	    rt->xm = (af + c3 * d) * f;
	    rt->xp = (sr * s3 * d) * f;
	} else {			/* q < 0 */
	    alpha = ASINH(ar / sq / aq);
	    c3 = sinh(alpha / 3.);
	    s3 = SQRT3 * cosh(alpha / 3.);
	    rt->x0 = (af - 2. * c3 * d) * f;
	    rt->xm = (af + c3 * d) * f;
	    rt->xp = (sr * s3 * d) * f;
	}
    }

    /* detect and flag degenerate roots */
    switch (rt->type) {
    case THREE:
	if (rt->x0 == rt->xm) {
	    if (rt->xm == rt->xp) {
		rt->type = TRIPLE;
	    } else {
		rt->type = LOWER_DOUBLE;
	    }
	} else if (rt->xm == rt->xp) {
	    rt->type = UPPER_DOUBLE;
	}
	break;
    case ONE:
	if (rt->xp == 0.) {
	    if (rt->x0 == rt->xm) {
		rt->type = TRIPLE;
		rt->xp = rt->xm;
	    } else if (rt->x0 < rt->xm) {
		rt->type = UPPER_DOUBLE;
		rt->xp = rt->xm;
	    } else {
		rt->type = LOWER_DOUBLE;
		rt->xp = rt->x0;
		rt->x0 = rt->xm;
	    }
	}
	break;
    }

    return(0);
}

#ifndef NOIEEE
/*------------------------------------------------------------------------
   Roots of cubic when any of the coefficients is NaN or +- Infinity
*/
int cubic_oo(a, b, c, rt)
REAL a, b, c;
roots3 *rt;
{
    int iret = 0;

    if (isnan(a)) {		/* a = NaN */
	rt->type = NONE;
	rt->x0 = rt->xm = rt->xp = a;
    } else if (isnan(b)) {	/* b = NaN */
	rt->type = NONE;
	rt->x0 = rt->xm = rt->xp = b;
    } else if (isnan(c)) {	/* c = NaN */
	rt->type = NONE;
	rt->x0 = rt->xm = rt->xp = c;
    } else if (isinf(a)) {
	if (a > 0.) {		/* a = Infinity */
	    rt->type = UPPER_DOUBLE;
	    rt->x0 = -a;
	    rt->xm = rt->xp = 0.;
	} else {		/* a = -Infinity */
	    rt->type = LOWER_DOUBLE;
	    rt->x0 = rt->xm = 0.;
	    rt->xp = -a;
	}
    } else if (isinf(b)) {
	if (b > 0.) {		/* b = Infinity */
	    rt->type = ONE;
	    rt->x0 = rt->xm = 0.;
	    rt->xp = b;
	} else {		/* b = -Infinity */
	    rt->type = THREE;
	    rt->x0 = b;
	    rt->xm = 0.;
	    rt->xp = -b;
	}
    } else if (isinf(c)) {	/* c = +- Infinity */
	rt->type = ONE;
	rt->x0 = -c;
	rt->xm = c;
	rt->xp = (c > 0.)? c : -c;
    } else {			/* a, b, c are all finite */
	iret = 1;
    }

    return(iret);
}
#endif
